# 
# Copyright (c) 1990-1996 The Regents of the University of California.
# All rights reserved.
# Permission is hereby granted, without written agreement and without
# license or royalty fees, to use, copy, modify, and distribute this
# software and its documentation for any purpose, provided that the
# above copyright notice and the following two paragraphs appear in all
# copies of this software.
# 
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
# ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
# THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# 
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
# PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
# CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
#
# 						PT_COPYRIGHT_VERSION_2
# 						COPYRIGHTENDKEY
#
#
# mkPtolemyTree: A shell script to build custom Ptolemy trees
#
# Version: @(#)mkPtolemyTree.tcl	1.17 3/3/96 
#
# Author: Jose Luis Pino
# Date: 10/20/95
#
# Port to Tcl by Matt Tavis and Brian L. Evans.
#
# We rely on 'mkdir -p' to create all the parent directories for
# a given path name.  The '-p' option is supported on the platforms
# we build for Ptolemy 0.6, and we use it in $PTOLEMY/src/kernel/Target.cc 
# and $PTOLEMY/src/domains/cg/kernel/CGUtilities.cc.  However, the
# '-p' option is not support on, for example, Ultrix 4.1.  See
# $PTOLEMY/src/gnu/src/make/make-stds.texi.  -cxh

proc replaceLinkWithDir {pathname linkspec} {
    if { [file type $pathname] == "link" } {
	puts "Creating $pathname"
	exec sh -c "rm $pathname ; mkdir -p $pathname ; ln -s $linkspec $pathname"
    } elseif { [file isdirectory $pathname] } {
	puts "The directory $pathname already exists"
    } else {
	puts "$pathname is neither a link nor a directory"
    }
}

proc processDirectory {mydir} {
    global CROOT ROOT
    set stddir "$mydir"
    regsub -all $CROOT "$stddir" $ROOT stddir
    puts "Expanding $mydir"
    set tarFiles ""
    set files [glob -nocomplain "$stddir/*"]
    foreach pathname $files {
	# extract the file name from the path name
	set file [file tail $pathname]

	# do not process junk files
	# -- skip core and TAGS files
	# -- skip files beginning with '#' or ',' or '.' and
	# -- skip files ending with ~
	if { [regexp {core|TAGS|(^#)|(^,)|(^\.)|(~$)} $file] } continue

	# if the file is not readable, then tell user and get next file
	if { ! [file readable "$pathname"] } {
	    puts "$pathname is not readable (skipped)"
	    continue
	}

	# if the file is a directory, then set a symbolic link to it;
	# otherwise, copy file unless it's executable (actually, we want
	# to avoid copying binaries but we do want to copy shell scripts;
	# however, Tcl does not have a convenient way of telling the
	# difference between an ascii and binary file).
	if { [file isdirectory "$stddir/$file"] } {
	    if {! [file exists "$mydir/$file"]} {		# prevent error
		exec ln -s "$stddir/$file" "$mydir/$file"
	    }
	} elseif {! [file executable "$stddir/$file"] } {
	    lappend tarFiles $file
	}
    }

    if { "$tarFiles" != "" } {
	exec sh -c "cd $stddir ; tar cf - $tarFiles | (cd $mydir ; tar xf -)"
    }
}

set error 0
if {$argc != 2} {
    puts "mkPtolemyTree: wrong number of arguments"
    set error 1
} elseif {! ([file readable [lindex $argv 0]] && \
	     ("file" == [file type [lindex $argv 0]]))} {
    puts "mkPtolemyTree: override.mk file unreadable"
    set error 1
} elseif {[file exists [lindex $argv 1]]} {
    puts "mkPtolemyTree: Custom root [lindex $argv 1] already exists"
    set error 1
} elseif {[string range [lindex $argv 1] 0 0] != "/"} {
    puts "mkPtolemyTree: [lindex $argv 1] must be expressed as an absolute path"
    set error 1
}    
if {$error} {
    puts "Usage: mkPtolemyTree <override.mk file> <root path of new tree>"
    exit
}

global env
set CROOT [lindex $argv 1]
set ROOT $env(PTOLEMY)
set OVERRIDE [lindex $argv 0]
set PTARCH $env(PTARCH)

# tell the user what is going on
puts "Making a customized Ptolemy development tree using the version of"
puts "Ptolemy installed in the directory $ROOT"
puts "The new customized Ptolemy tree will go in $CROOT"

# here are some libraries which are in lib.$PTACH but do not get built often...
# so, we just make a sym link to them
set LIBDIR "$ROOT/lib.$PTARCH"
set AUX_LIBRARIES "$LIBDIR/libCGCrtlib.* $LIBDIR/libcephes.*"

### The next few lines are all part of MAKEFILE value
set MAKEFILE "
TK=1
OCT=1
ROOT=$ROOT
CROOT=$CROOT
PTARCH=$PTARCH
include $OVERRIDE
include $ROOT/mk/stars.mk

TREE:
	echo -n \$(CUSTOM_DIRS)
"
### Here ends the setting of MAKEFILE value

set srcdirs [exec echo "$MAKEFILE" | make -s -f - ]

set objdirs $srcdirs
regsub -all "/src/" $objdirs "/obj.$PTARCH/" objdirs

foreach mdir $objdirs {
	exec mkdir -p "$mdir"
}
foreach mdir $srcdirs {
	exec mkdir -p "$mdir"
}

# process each directory, src directories before obj directories so that
# the make.template and makefile symlinks are done correctly
foreach dir [exec find $CROOT -type d -print | sort -r] {
    set dir [string trim $dir]
    if { [regexp "/obj\.$PTARCH/" $dir] } {
	set srcdir $dir
	regsub -all "/obj\.$PTARCH/" $srcdir "/src/" srcdir
	if { [file exists "$srcdir/makefile"] } {
	    exec ln -s "$srcdir/makefile" "$dir/makefile"
	}
	if { [file exists "$srcdir/make.template"] } {
	    exec ln -s "$srcdir/make.template" "$dir/make.template"
	}
    }
    processDirectory $dir
}

puts "Copying $OVERRIDE to $CROOT/mk/override.mk"
exec cp $OVERRIDE $CROOT/mk/override.mk

set OVER [open "$CROOT/obj.$PTARCH/pigiRpc/override.mk" w]
puts $OVER "include \$(ROOT)/mk/override.mk\nPTOLEMY=$CROOT\nPIGI=pigiRpc\n"
close $OVER

set OVER [open "$CROOT/obj.$PTARCH/ptcl/override.mk" w]
puts $OVER "include \$(ROOT)/mk/override.mk\nPTOLEMY=$CROOT\nPIGI=ptcl\n"
close $OVER

set OVER [open "$CROOT/obj.$PTARCH/tycho/tysh/override.mk" w]
puts $OVER "include \$(ROOT)/mk/override.mk\nPTOLEMY=$CROOT\nPIGI=tysh\n"
close $OVER

# Remove the symbolic links for bin, bin.$PTARCH, and lib.$PTARCH,
# create these directories, and set up initial symbolic links
replaceLinkWithDir "$CROOT/bin" "$ROOT/bin/*"
replaceLinkWithDir "$CROOT/bin.$PTARCH" "$ROOT/bin.$PTARCH/*"
replaceLinkWithDir "$CROOT/lib" "$ROOT/lib/*"
replaceLinkWithDir "$CROOT/lib.$PTARCH" "$AUX_LIBRARIES"

# We need to make sources to build any makefiles that are out of date
# and to make install to compile any out-of-date object files
puts "Building ptcl, pigiRpc, and tycho in your customized Ptolemy tree."
puts "Please wait for a few minutes ..."
exec sh -c "cd $CROOT/obj.$PTARCH ; make sources install"
